﻿namespace TabsStudioPriorityGroup
{
    class Controller
    {
        public Controller(TabsStudioExt.ITabsStudioEngine engine_)
        {
            engine = engine_;
            openTabs = new System.Collections.Generic.List<TabInfo>();
            storage = new OptionsStorage(engine.GetAddinSettingsDirectory());
            options = storage.Load();
            isPreviewTabProperty = new TabsStudioExt.Util.IsPreviewTabProperty();
            tabNameProperty = new TabsStudioExt.Util.TabNameProperty();
            superGroupProperties = new TabsStudioExt.Util.SuperGroupProperties();
        }

        public void OnTabCreated(TabsStudioExt.TabEventArgs e)
        {
            TabInfo info = new TabInfo(e.Tab.TabItem,
                (object o, System.EventArgs a) =>
                    { SortTabAndUpdateProperties(e.Tab.TabItem, e.Tabs.TabPanel2); });
            openTabs.Add(info);
            tabNameProperty.GetDescriptor(e.Tab.TabItem).AddValueChanged(info.item, info.tabNameChanged);
            SortTabAndUpdateProperties(e.Tab.TabItem, e.Tabs.TabPanel2);
        }

        public void OnTabDestroyed(TabsStudioExt.TabEventArgs e)
        {
            foreach (TabInfo info in openTabs)
            {
                if (info.item == e.Tab.TabItem)
                {
                    tabNameProperty.GetDescriptor(e.Tab.TabItem).RemoveValueChanged(info.item, info.tabNameChanged);
                    openTabs.Remove(info);
                    break;
                }
            }
            UpdateDeletedSuperGroupProperties(e.Tab.TabItem, e.Tabs.TabPanel2);
        }

        public void AddTabToPriorityGroup(TabsStudioExt.Tab tab)
        {
            PriorityRule rule = new PriorityRule();
            rule.tabNameFilter = GetTabName(tab.TabItem);
            options.priorityRules.Add(rule);
            UpdateRules();
        }

        public void RemoveTabFromPriorityGroup(TabsStudioExt.Tab tab)
        {
            string name = GetTabName(tab.TabItem);
            foreach (PriorityRule rule in options.priorityRules)
            {
                if (rule.IsMatch(name))
                {
                    options.priorityRules.Remove(rule);
                    break;
                }
            }
            UpdateRules();
        }

        public bool IsPriorityTab(TabsStudioExt.Tab tab)
        {
            return IsPriorityTabItem(tab.TabItem);
        }

        private void UpdateRules()
        {
            storage.Save(options);

            foreach (TabsStudioExt.Tabs tabs in engine.GetTabs())
            {
                SortTabs(tabs);
                UpdateProperties(tabs);
            }
        }

        private void SortTabs(TabsStudioExt.Tabs tabs)
        {
            System.Windows.Controls.Panel panel = tabs.TabPanel2;
            for (int a = 0; a < panel.Children.Count; ++a)
            {
                System.Windows.Controls.TabItem tabItem = TabItemChild(panel, a);
                if (IsPriorityTabItem(tabItem))
                {
                    for (int i = 0; i < panel.Children.IndexOf(tabItem); ++i)
                    {
                        if (!IsPriorityTabItem(TabItemChild(panel, i)))
                        {
                            panel.Children.Remove(tabItem);
                            panel.Children.Insert(i, tabItem);
                            break;
                        }
                    }
                }
            }
        }

        private void UpdateProperties(TabsStudioExt.Tabs tabs)
        {
            foreach (TabsStudioExt.Tab tab in tabs.TabList)
                Properties.SetTop(tab.TabItem, IsPriorityTabItem(tab.TabItem));
            UpdateSuperGroupProperties(tabs.TabPanel2);
        }

        private void SortTabAndUpdateProperties(System.Windows.Controls.TabItem tabItem, System.Windows.Controls.Panel panel)
        {
            if (isPreviewTabProperty.IsPreviewTab(tabItem))
                return;

            bool isPriority = IsPriorityTabItem(tabItem);
            if (isPriority)
            {
                for (int i = 0; i < panel.Children.IndexOf(tabItem); ++i)
                {
                    if (!IsPriorityTabItem(TabItemChild(panel, i)))
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(i, tabItem);
                        break;
                    }
                }
            }
            else
            {
                for (int i = panel.Children.Count - 1; i > panel.Children.IndexOf(tabItem); --i)
                {
                    if (IsPriorityTabItem(TabItemChild(panel, i)))
                    {
                        panel.Children.Remove(tabItem);
                        panel.Children.Insert(i, tabItem);
                        break;
                    }
                }
            }
            Properties.SetTop(tabItem, IsPriorityTabItem(tabItem));
            UpdateSuperGroupProperties(panel);
        }

        private bool IsPriorityTabItem(System.Windows.Controls.TabItem tabItem)
        {
            string name = GetTabName(tabItem);
            if (name == null)
                return false;

            foreach (PriorityRule rule in options.priorityRules)
            {
                if (rule.IsMatch(name))
                    return true;
            }
            return false;
        }

        private string GetTabName(System.Windows.Controls.TabItem tabItem)
        {
            return tabNameProperty.GetName(tabItem);
        }

        private void UpdateSuperGroupProperties(System.Windows.Controls.Panel panel)
        {
            foreach (System.Windows.Controls.TabItem i in panel.Children)
            {
                if (IsPriorityTabItem(i))
                {
                    superGroupProperties.SetName(i, superGroupName);
                }
                else if (superGroupProperties.GetName(i) == superGroupName)
                {
                    superGroupProperties.SetName(i, null);
                    superGroupProperties.SetFirstTab(i, false);
                    superGroupProperties.SetLastTab(i, false);
                }
            }

            for (int i = 0; i < panel.Children.Count; ++i)
            {
                System.Windows.Controls.TabItem tab = TabItemChild(panel, i);
                if (IsPriorityTabItem(tab))
                {
                    superGroupProperties.SetFirstTab(tab,
                        i == 0 || !IsPriorityTabItem(TabItemChild(panel, i - 1)));
                    superGroupProperties.SetLastTab(tab,
                        i == panel.Children.Count - 1 || !IsPriorityTabItem(TabItemChild(panel, i + 1)));
                }
            }
        }

        private void UpdateDeletedSuperGroupProperties(System.Windows.Controls.TabItem deletedTabItem,
            System.Windows.Controls.Panel panel)
        {
            if (superGroupProperties.IsFirstTab(deletedTabItem))
            {
                int nextTabIndex = panel.Children.IndexOf(deletedTabItem) + 1;
                if (nextTabIndex < panel.Children.Count)
                {
                    System.Windows.Controls.TabItem tab = TabItemChild(panel, nextTabIndex);
                    if (IsPriorityTabItem(tab))
                        superGroupProperties.SetFirstTab(tab, true);
                }
            }
            if (superGroupProperties.IsLastTab(deletedTabItem))
            {
                int prevTabIndex = panel.Children.IndexOf(deletedTabItem) - 1;
                if (prevTabIndex >= 0)
                {
                    System.Windows.Controls.TabItem tab = TabItemChild(panel, prevTabIndex);
                    if (IsPriorityTabItem(tab))
                        superGroupProperties.SetLastTab(tab, true);
                }
            }
        }

        private System.Windows.Controls.TabItem TabItemChild(System.Windows.Controls.Panel panel, int i)
        {
            return (System.Windows.Controls.TabItem)panel.Children[i];
        }

        private TabsStudioExt.ITabsStudioEngine engine;
        private System.Collections.Generic.List<TabInfo> openTabs;
        private OptionsStorage storage;
        private Options options;
        private TabsStudioExt.Util.IsPreviewTabProperty isPreviewTabProperty;
        private TabsStudioExt.Util.TabNameProperty tabNameProperty;
        private TabsStudioExt.Util.SuperGroupProperties superGroupProperties;
        private readonly string superGroupName = "[PriorityGroup]";
    }
}
